@@ -10,7 +10,7 @@
  * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
  * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
  *
- * $Id: wl_linux.c,v 1.480.2.12.2.14.8.1 2010/09/29 17:56:48 Exp $
+ * $Id: wl_linux.c,v 1.524.2.40.2.12 2010/12/08 23:33:57 Exp $
  */
 
 #define LINUX_PORT
@@ -39,9 +39,7 @@
 #include <linux/completion.h>
 #include <linux/pci_ids.h>
 #define WLC_MAXBSSCFG		1	
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)
-#include <linux/sched.h>
-#endif
+
 #include <asm/system.h>
 #include <asm/io.h>
 #include <asm/irq.h>
@@ -61,9 +59,6 @@
 
 typedef const struct si_pub	si_t;
 
-typedef void wlc_info_t;
-typedef void wlc_hw_info_t;
-
 #if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 4, 5)
 #error "No support for Kernel Rev <= 2.4.5, As the older kernel revs doesn't support Tasklets"
 #endif
@@ -71,18 +66,39 @@ typedef void wlc_hw_info_t;
 #include <wlc_pub.h>
 #include <wl_dbg.h>
 
-#ifdef CONFIG_WIRELESS_EXT
+#undef USE_IW
+#if defined(CONFIG_WIRELESS_EXT)
+#define USE_IW
+#endif
+#ifdef USE_IW
 #include <wl_iw.h>
 struct iw_statistics *wl_get_wireless_stats(struct net_device *dev);
-#endif 
+#endif
 
 #include <wl_export.h>
 
 #include <wl_linux.h>
 
+#ifdef WL_THREAD
+#include <linux/kthread.h>
+#endif 
 static void wl_timer(ulong data);
 static void _wl_timer(wl_timer_t *t);
 
+#ifdef WL_ALL_PASSIVE
+static void wl_start_txqwork(struct wl_task *task);
+static void wl_txq_free(wl_info_t *wl);
+#define TXQ_LOCK(_wl)	spin_lock_bh(&(_wl)->txq_lock)
+#define TXQ_UNLOCK(_wl)	spin_unlock_bh(&(_wl)->txq_lock)
+
+static void wl_set_multicast_list_workitem(struct work_struct *work);
+
+static void wl_timer_task(wl_task_t *task);
+static void wl_dpc_rxwork(struct wl_task *task);
+#else
+
+#endif 
+
 static int wl_linux_watchdog(void *ctx);
 static
 int wl_found = 0;
@@ -120,7 +136,11 @@ struct ieee80211_tkip_data {
 
 static int wl_open(struct net_device *dev);
 static int wl_close(struct net_device *dev);
-static int wl_start(struct sk_buff *skb, struct net_device *dev);
+#ifdef WL_THREAD
+static int wl_start_wlthread(struct sk_buff *skb, struct net_device *dev);
+#else
+static int BCMFASTPATH wl_start(struct sk_buff *skb, struct net_device *dev);
+#endif
 static int wl_start_int(wl_info_t *wl, wl_if_t *wlif, struct sk_buff *skb);
 
 static struct net_device_stats *wl_get_stats(struct net_device *dev);
@@ -128,12 +148,20 @@ static int wl_set_mac_address(struct net_device *dev, void *addr);
 static void wl_set_multicast_list(struct net_device *dev);
 static void _wl_set_multicast_list(struct net_device *dev);
 static int wl_ethtool(wl_info_t *wl, void *uaddr, wl_if_t *wlif);
+#ifdef NAPI_POLL
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
+static int wl_poll(struct napi_struct *napi, int budget);
+#else
+static int wl_poll(struct net_device *dev, int *budget);
+#endif 
+#else 
 static void wl_dpc(ulong data);
+#endif 
 static void wl_link_up(wl_info_t *wl, char * ifname);
 static void wl_link_down(wl_info_t *wl, char *ifname);
-#if defined(WL_CONFIG_RFKILL_INPUT)
-static int wl_schedule_task(wl_info_t *wl, void (*fn)(struct wl_task *), void *context);
-#endif
+#ifdef WL_THREAD
+static int wl_start_enqueue_wlthread(wl_info_t *wl, struct sk_buff *skb);
+#endif 
 #if defined(CONFIG_PROC_FS)
 static int wl_read_proc(char *buffer, char **start, off_t offset, int length, int *eof, void *data);
 #endif 
@@ -144,11 +172,12 @@ static struct wl_if *wl_alloc_if(wl_info_t *wl, int iftype, uint unit, struct wl
 static void wl_free_if(wl_info_t *wl, wl_if_t *wlif);
 static void wl_get_driver_info(struct net_device *dev, struct ethtool_drvinfo *info);
 
-#if defined(WL_CONFIG_RFKILL_INPUT)
+#if defined(WL_CONFIG_RFKILL)
 #include <linux/rfkill.h>
 static int wl_init_rfkill(wl_info_t *wl);
 static void wl_uninit_rfkill(wl_info_t *wl);
-static void wl_force_kill(wl_task_t *task);
+static int wl_set_radio_block(void *data, bool blocked);
+static void wl_report_radio_state(wl_info_t *wl);
 #endif
 
 static struct pci_device_id wl_id_table[] = {
@@ -163,8 +192,13 @@ static struct pci_device_id wl_id_table[] = {
 	{ PCI_VENDOR_ID_BROADCOM, 0x432c, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 }, 
 	{ PCI_VENDOR_ID_BROADCOM, 0x432d, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 }, 
 	{ PCI_VENDOR_ID_BROADCOM, 0x4353, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 }, 
+	{ PCI_VENDOR_ID_BROADCOM, 0xA99D, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 }, 
 	{ PCI_VENDOR_ID_BROADCOM, 0x4357, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 }, 
 	{ PCI_VENDOR_ID_BROADCOM, 0x4727, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 }, 
+	{ PCI_VENDOR_ID_BROADCOM, 0x4358, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 }, 
+	{ PCI_VENDOR_ID_BROADCOM, 0x4359, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 }, 
+	{ PCI_VENDOR_ID_BROADCOM, 0x435a, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 }, 
+	{ PCI_VENDOR_ID_BROADCOM, 0x0576, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 }, 
 	{ 0 }
 };
 MODULE_DEVICE_TABLE(pci, wl_id_table);
@@ -174,6 +208,19 @@ static int msglevel = 0xdeadbeef;
 module_param(msglevel, int, 0);
 static int msglevel2 = 0xdeadbeef;
 module_param(msglevel2, int, 0);
+static int phymsglevel = 0xdeadbeef;
+module_param(phymsglevel, int, 0);
+#endif 
+
+#if defined(WL_ALL_PASSIVE)
+
+#ifdef WLP2P
+static int passivemode = 1;
+module_param(passivemode, int, 1);
+#else
+static int passivemode = 0;
+module_param(passivemode, int, 0);
+#endif 
 #endif 
 
 static int oneonly = 0;
@@ -185,17 +232,47 @@ module_param(piomode, int, 0);
 static int instance_base = 0;	
 module_param(instance_base, int, 0);
 
+#if defined(BCMDBG)
+static struct ether_addr local_ea;
+static char *macaddr = NULL;
+module_param(macaddr, charp, S_IRUGO);
+#endif
+
 static int nompc = 0;
 module_param(nompc, int, 0);
 
-static char name[IFNAMSIZ] = "eth%d";
+#ifdef quote_str
+#undef quote_str
+#endif 
+#ifdef to_str
+#undef to_str
+#endif 
+#define to_str(s) #s
+#define quote_str(s) to_str(s)
+
+#ifndef BRCM_WLAN_IFNAME
+#define BRCM_WLAN_IFNAME eth%d
+#endif
+
+static char name[IFNAMSIZ] = quote_str(BRCM_WLAN_IFNAME);
+
 module_param_string(name, name, IFNAMSIZ, 0);
 
+#define WL_RADIOTAP_BRCM_SNS	0x01
+#define WL_RADIOTAP_BRCM_MCS	0x00000001
+
+#define	IEEE80211_RADIOTAP_HTMOD_40		0x01
+#define	IEEE80211_RADIOTAP_HTMOD_SGI		0x02
+#define	IEEE80211_RADIOTAP_HTMOD_GF		0x04
+#define	IEEE80211_RADIOTAP_HTMOD_LDPC		0x08
+#define	IEEE80211_RADIOTAP_HTMOD_STBC_MASK	0x30
+#define	IEEE80211_RADIOTAP_HTMOD_STBC_SHIFT	4
+
 #ifndef	SRCBASE
 #define	SRCBASE "."
-#endif 
+#endif
 
-#if WIRELESS_EXT >= 19
+#if WIRELESS_EXT >= 19 || LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
 static struct ethtool_ops wl_ethtool_ops =
 #else
@@ -205,18 +282,51 @@ static const struct ethtool_ops wl_ethtool_ops =
 	.get_drvinfo = wl_get_driver_info
 };
 #endif 
+#ifdef WL_THREAD
+static int wl_thread_dpc_wlthread(void *data)
+{
+	wl_info_t *wl = (wl_info_t *) data;
+
+	current->flags |= PF_NOFREEZE;
+
+	while (1) {
+		wait_event_interruptible_timeout
+		(wl->thread_wqh,
+			skb_queue_len(&wl->rpc_queue) ||
+			skb_queue_len(&wl->tx_queue),
+			1);
+
+		if (kthread_should_stop())
+			break;
+
+		wl_rpcq_dispatch_wlthread(wl);
+		wl_start_txqwork_wlthread(wl);
+	}
+
+	skb_queue_purge(&wl->tx_queue);
+	skb_queue_purge(&wl->rpc_queue);
+
+	return 0;
+}
+#endif 
 
 #if defined(WL_USE_NETDEV_OPS)
+
 static const struct net_device_ops wl_netdev_ops =
 {
 	.ndo_open = wl_open,
 	.ndo_stop = wl_close,
+#ifdef WL_THREAD
+	.ndo_start_xmit = wl_start_wlthread,
+#else
 	.ndo_start_xmit = wl_start,
+#endif
 	.ndo_get_stats = wl_get_stats,
 	.ndo_set_mac_address = wl_set_mac_address,
 	.ndo_set_multicast_list = wl_set_multicast_list,
 	.ndo_do_ioctl = wl_ioctl
 };
+
 #endif 
 
 static
@@ -227,26 +337,45 @@ void wl_if_setup(struct net_device *dev)
 #else
 	dev->open = wl_open;
 	dev->stop = wl_close;
+#ifdef WL_THREAD
+	dev->hard_start_xmit = wl_start_wlthread;
+#else
 	dev->hard_start_xmit = wl_start;
+#endif
 	dev->get_stats = wl_get_stats;
 	dev->set_mac_address = wl_set_mac_address;
 	dev->set_multicast_list = wl_set_multicast_list;
 	dev->do_ioctl = wl_ioctl;
 #endif 
+#ifdef NAPI_POLL
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
+	{
+		struct wl_info *wl = WL_INFO(dev);
 
-#ifdef CONFIG_WIRELESS_EXT
+		netif_napi_add(dev, &wl->napi, wl_poll, 64);
+		napi_enable(&wl->napi);
+	}
+#else
+	dev->poll = wl_poll;
+	dev->weight = 64;
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 21)
+	netif_poll_enable(dev);
+#endif 
+#endif  
+#endif 
+
+#ifdef USE_IW
 #if WIRELESS_EXT < 19
 	dev->get_wireless_stats = wl_get_wireless_stats;
-#endif 
+#endif
 #if WIRELESS_EXT > 12
 	dev->wireless_handlers = (struct iw_handler_def *) &wl_iw_handler_def;
+#endif
 #endif 
-#if WIRELESS_EXT >= 19
+
+#if WIRELESS_EXT >= 19 || LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
 	dev->ethtool_ops = &wl_ethtool_ops;
-#endif 
-
-#endif 
-
+#endif
 }
 
 static wl_info_t *
@@ -259,9 +388,10 @@ wl_attach(uint16 vendor, uint16 device, ulong regs, uint bustype, void *btparam,
 	char tmp[128];
 #endif
 	osl_t *osh;
-	int unit;
+	int unit, err;
 
 	unit = wl_found + instance_base;
+	err = 0;
 
 	if (unit < 0) {
 		WL_ERROR(("wl%d: unit number overflow, exiting\n", unit));
@@ -285,8 +415,25 @@ wl_attach(uint16 vendor, uint16 device, ulong regs, uint bustype, void *btparam,
 	bzero(wl, sizeof(wl_info_t));
 
 	wl->osh = osh;
+	wl->unit = unit;
 	atomic_set(&wl->callbacks, 0);
 
+#ifdef WL_ALL_PASSIVE
+	wl->all_dispatch_mode = (passivemode == 0) ? TRUE : FALSE;
+	if (WL_ALL_PASSIVE_ENAB(wl)) {
+
+		MY_INIT_WORK(&wl->txq_task.work, (work_func_t)wl_start_txqwork);
+		wl->txq_task.context = wl;
+		wl->txq_dispatched = FALSE;
+		wl->txq_head = wl->txq_tail = NULL;
+
+		MY_INIT_WORK(&wl->multicast_task.work, (work_func_t)wl_set_multicast_list_workitem);
+
+		MY_INIT_WORK(&wl->wl_dpc_task.work, (work_func_t)wl_dpc_rxwork);
+		wl->wl_dpc_task.context = wl;
+	}
+#endif  
+
 	wlif = wl_alloc_if(wl, WL_IFTYPE_BSS, unit, NULL);
 	if (!wlif) {
 		WL_ERROR(("wl%d: wl_alloc_if failed\n", unit));
@@ -331,14 +478,18 @@ wl_attach(uint16 vendor, uint16 device, ulong regs, uint bustype, void *btparam,
 	spin_lock_init(&wl->lock);
 	spin_lock_init(&wl->isr_lock);
 
-	{
-	int err;
+	if (WL_ALL_PASSIVE_ENAB(wl)) {
+#ifdef WL_ALL_PASSIVE
+		spin_lock_init(&wl->txq_lock);
+#endif 
+		init_MUTEX(&wl->sem);
+	}
+
 	if (!(wl->wlc = wlc_attach((void *) wl, vendor, device, unit, wl->piomode,
 		osh, wl->regsva, wl->bcm_bustype, btparam, &err))) {
 		printf("%s: %s driver failed with code %d\n", dev->name, EPI_VERSION_STR, err);
 		goto fail;
 	}
-	}
 	wl->pub = wlc_pub(wl->wlc);
 
 	if (nompc) {
@@ -355,11 +506,27 @@ wl_attach(uint16 vendor, uint16 device, ulong regs, uint bustype, void *btparam,
 
 	sprintf(tmp, "net/wl%d", wl->pub->unit);
 	create_proc_read_entry(tmp, 0, 0, wl_read_proc, (void*)wl);
-#endif 
+#endif
 
+#ifdef BCMDBG
+	if (macaddr != NULL) {  
+		int err;
+
+		WL_ERROR(("wl%d: setting MAC ADDRESS %s\n", unit, macaddr));
+		bcm_ether_atoe(macaddr, &local_ea);
+
+		err = wlc_iovar_op(wl->wlc, "cur_etheraddr", NULL, 0, &local_ea,
+			ETHER_ADDR_LEN, IOV_SET, NULL);
+		if (err)
+			WL_ERROR(("wl%d: Error setting MAC ADDRESS\n", unit));
+	}
+#endif	
 	bcopy(&wl->pub->cur_etheraddr, dev->dev_addr, ETHER_ADDR_LEN);
 
+#ifndef NAPI_POLL
+
 	tasklet_init(&wl->tasklet, wl_dpc, (ulong)wl);
+#endif
 
 	{
 		if (request_irq(irq, wl_isr, IRQF_SHARED, dev->name, wl)) {
@@ -397,11 +564,11 @@ wl_attach(uint16 vendor, uint16 device, ulong regs, uint bustype, void *btparam,
 	}
 #endif 
 #endif 
-#ifdef CONFIG_WIRELESS_EXT
+#ifdef USE_IW
 	wlif->iw.wlinfo = (void *)wl;
 #endif
 
-#if defined(WL_CONFIG_RFKILL_INPUT)
+#if defined(WL_CONFIG_RFKILL)
 	if (wl_init_rfkill(wl) < 0)
 		WL_ERROR(("%s: init_rfkill_failure\n", __FUNCTION__));
 #endif
@@ -421,6 +588,22 @@ wl_attach(uint16 vendor, uint16 device, ulong regs, uint bustype, void *btparam,
 		WL_ERROR(("wl%d: Error setting infra_mode to infrastructure\n", unit));
 	}
 
+#ifdef DEFAULT_EAPVER_AP
+
+	if (wlc_iovar_setint(wl->wlc, "sup_wpa2_eapver", -1)) {
+		WL_ERROR(("wl%d: Error setting sup_wpa2_eapver \n", unit));
+	}
+	if (wlc_iovar_setint(wl->wlc, "sup_m3sec_ok", 1)) {
+		WL_ERROR(("wl%d: Error setting sup_m3sec_ok \n", unit));
+	}
+#endif 
+#ifdef DISABLE_HT_RATE_FOR_WEP_TKIP
+
+	if (wlc_iovar_setint(wl->wlc, "ht_wsec_restrict", 0x3)) {
+		WL_ERROR(("wl%d: Error setting ht_wsec_restrict \n", unit));
+	}
+#endif 
+
 	wlc_module_register(wl->pub, NULL, "linux", wl, NULL, wl_linux_watchdog, NULL);
 
 #ifdef BCMDBG
@@ -496,8 +679,13 @@ wl_pci_probe(struct pci_dev *pdev, const struct pci_device_id *ent)
 
 	if ((pdev->vendor != PCI_VENDOR_ID_BROADCOM) ||
 	    (((pdev->device & 0xff00) != 0x4300) &&
-	     ((pdev->device & 0xff00) != 0x4700)))
+		(pdev->device != 0x576) &&
+	     ((pdev->device & 0xff00) != 0x4700) &&
+	     ((pdev->device < 43000) || (pdev->device > 43999)))) {
+		WL_TRACE(("%s: unsupported vendor %x device %x\n", __FUNCTION__,
+		          pdev->vendor, pdev->device));
 		return (-ENODEV);
+	}
 
 	rc = pci_enable_device(pdev);
 	if (rc) {
@@ -615,6 +803,10 @@ static struct pci_driver wl_pci_driver = {
 	id_table:	wl_id_table,
 	};
 
+#define	wl_read_nvram_file() do {} while (0)
+
+#define usbap_parse_nvram()	do {} while (0)
+
 static int __init
 wl_module_init(void)
 {
@@ -637,6 +829,26 @@ wl_module_init(void)
 			wl_msg_level2 = bcm_strtoul(var, NULL, 0);
 	}
 	printf("%s: msglevel2 set to 0x%x\n", __FUNCTION__, wl_msg_level2);
+	{
+		extern uint32 phyhal_msg_level;
+
+		if (phymsglevel != 0xdeadbeef)
+			phyhal_msg_level = phymsglevel;
+		else {
+			char *var = getvar(NULL, "phy_msglevel");
+			if (var)
+				phyhal_msg_level = bcm_strtoul(var, NULL, 0);
+		}
+		printf("%s: phymsglevel set to 0x%x\n", __FUNCTION__, phyhal_msg_level);
+	}
+#if defined(WL_ALL_PASSIVE)
+	{
+		char *var = getvar(NULL, "wl_dispatch_mode");
+		if (var)
+			passivemode = bcm_strtoul(var, NULL, 0);
+		printf("%s: dhssivemode set to 0x%x\n", __FUNCTION__, passivemode);
+	}
+#endif 
 #endif 
 
 	if (!(error = pci_module_init(&wl_pci_driver)))
@@ -668,17 +880,29 @@ wl_free(wl_info_t *wl)
 			free_irq(wl->dev->irq, wl);
 	}
 
-#if defined(WL_CONFIG_RFKILL_INPUT)
+#if defined(WL_CONFIG_RFKILL)
 	wl_uninit_rfkill(wl);
 #endif
 
+#ifdef NAPI_POLL
+	clear_bit(__LINK_STATE_START, &wl->dev->state);
+
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
+	napi_disable(&wl->napi);
+#elif  LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 21)
+	netif_poll_disable(wl->dev);
+#endif 	
+#endif 
+
 	if (wl->dev) {
 		wl_free_if(wl, WL_DEV_IF(wl->dev));
 		wl->dev = NULL;
 	}
 
-	tasklet_kill(&wl->tasklet);
+#ifndef NAPI_POLL
 
+	tasklet_kill(&wl->tasklet);
+#endif
 	if (wl->pub) {
 		wlc_module_unregister(wl->pub, "linux", wl);
 	}
@@ -707,9 +931,9 @@ wl_free(wl_info_t *wl)
 		MFREE(wl->osh, t, sizeof(wl_timer_t));
 	}
 
-	if (wl->monitor) {
-		wl_free_if(wl, WL_DEV_IF(wl->monitor));
-		wl->monitor = NULL;
+	if (wl->monitor_dev) {
+		wl_free_if(wl, WL_DEV_IF(wl->monitor_dev));
+		wl->monitor_dev = NULL;
 	}
 
 	osh = wl->osh;
@@ -737,6 +961,11 @@ wl_free(wl_info_t *wl)
 	}
 #endif 
 
+#ifdef WL_ALL_PASSIVE
+	if (WL_ALL_PASSIVE_ENAB(wl))
+		wl_txq_free(wl);
+#endif 
+
 	MFREE(osh, wl, sizeof(wl_info_t));
 
 	if (MALLOCED(osh)) {
@@ -799,15 +1028,6 @@ wl_close(struct net_device *dev)
 	return (0);
 }
 
-static int BCMFASTPATH
-wl_start(struct sk_buff *skb, struct net_device *dev)
-{
-	if (!dev)
-		return -ENETDOWN;
-
-	return wl_start_int(WL_INFO(dev), WL_DEV_IF(dev), skb);
-}
-
 static int BCMFASTPATH
 wl_start_int(wl_info_t *wl, wl_if_t *wlif, struct sk_buff *skb)
 {
@@ -836,20 +1056,24 @@ wl_start_int(wl_info_t *wl, wl_if_t *wlif, struct sk_buff *skb)
 }
 
 void
-wl_txflowcontrol(wl_info_t *wl, bool state, int prio)
+wl_txflowcontrol(wl_info_t *wl, struct wl_if *wlif, bool state, int prio)
 {
-	wl_if_t *wlif;
+	struct net_device *dev;
 
 	ASSERT(prio == ALLPRIO);
-	for (wlif = wl->if_list; wlif != NULL; wlif = wlif->next) {
-		if (state == ON)
-			netif_stop_queue(wlif->dev);
-		else
-			netif_wake_queue(wlif->dev);
-	}
+
+	if (wlif == NULL)
+		dev = wl->dev;
+	else
+		dev = wlif->dev;
+
+	if (state == ON)
+		netif_stop_queue(dev);
+	else
+		netif_wake_queue(dev);
 }
 
-#if defined(WL_CONFIG_RFKILL_INPUT)
+#if defined(WL_ALL_PASSIVE)
 
 static int
 wl_schedule_task(wl_info_t *wl, void (*fn)(struct wl_task *task), void *context)
@@ -894,6 +1118,7 @@ wl_alloc_if(wl_info_t *wl, int iftype, uint subunit, struct wlc_if* wlcif)
 	}
 	bzero(wlif, sizeof(wl_if_t));
 #endif 
+
 #if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 24))
 	if (!(dev = MALLOC(wl->osh, sizeof(struct net_device)))) {
 		MFREE(wl->osh, wlif, sizeof(wl_if_t));
@@ -905,19 +1130,19 @@ wl_alloc_if(wl_info_t *wl, int iftype, uint subunit, struct wlc_if* wlcif)
 	ether_setup(dev);
 	strncpy(dev->name, name, IFNAMSIZ);
 #else
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
+#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29))
 
 	dev = alloc_netdev(sizeof(wl_if_t), name, ether_setup);
 	wlif = netdev_priv(dev);
 	if (!dev) {
 #else
-		dev = alloc_netdev(0, name, ether_setup);
-		if (!dev) {
-			MFREE(wl->osh, wlif, sizeof(wl_if_t));
-#endif
-			WL_ERROR(("wl%d: wl_alloc_if: out of memory, alloc_netdev\n",
-				(wl->pub)?wl->pub->unit:subunit));
-			return NULL;
+	dev = alloc_netdev(0, name, ether_setup);
+	if (!dev) {
+		MFREE(wl->osh, wlif, sizeof(wl_if_t));
+#endif 
+		WL_ERROR(("wl%d: wl_alloc_if: out of memory, alloc_netdev\n",
+			(wl->pub)?wl->pub->unit:subunit));
+		return NULL;
 	}
 #endif 
 
@@ -948,8 +1173,9 @@ wl_free_if(wl_info_t *wl, wl_if_t *wlif)
 {
 	wl_if_t *p;
 
-	if (wlif->dev_registed)
+	if (wlif->dev_registed) {
 		unregister_netdev(wlif->dev);
+	}
 
 	p = wl->if_list;
 	if (p == wlif)
@@ -1008,7 +1234,7 @@ wl_reset(wl_info_t *wl)
 void BCMFASTPATH
 wl_intrson(wl_info_t *wl)
 {
-	unsigned long flags;
+	unsigned long flags = 0;
 
 	INT_LOCK(wl, flags);
 	wlc_intrson(wl->wlc);
@@ -1024,7 +1250,7 @@ wl_alloc_dma_resources(wl_info_t *wl, uint addrwidth)
 uint32 BCMFASTPATH
 wl_intrsoff(wl_info_t *wl)
 {
-	unsigned long flags;
+	unsigned long flags = 0;
 	uint32 status;
 
 	INT_LOCK(wl, flags);
@@ -1036,7 +1262,7 @@ wl_intrsoff(wl_info_t *wl)
 void
 wl_intrsrestore(wl_info_t *wl, uint32 macintmask)
 {
-	unsigned long flags;
+	unsigned long flags = 0;
 
 	INT_LOCK(wl, flags);
 	wlc_intrsrestore(wl->wlc, macintmask);
@@ -1055,8 +1281,17 @@ wl_up(wl_info_t *wl)
 
 	error = wlc_up(wl->wlc);
 
-	if (!error)
-		wl_txflowcontrol(wl, OFF, ALLPRIO);
+	if (!error) {
+		wl_if_t *wlif;
+
+		for (wlif = wl->if_list; wlif != NULL; wlif = wlif->next) {
+			wl_txflowcontrol(wl, wlif, OFF, ALLPRIO);
+		}
+	}
+
+#ifdef NAPI_POLL
+	set_bit(__LINK_STATE_START, &wl->dev->state);
+#endif 
 
 	return (error);
 }
@@ -1079,7 +1314,20 @@ wl_down(wl_info_t *wl)
 
 	WL_UNLOCK(wl);
 
-	SPINWAIT((atomic_read(&wl->callbacks) > callbacks), 100 * 1000);
+#ifdef WL_ALL_PASSIVE
+	if (WL_ALL_PASSIVE_ENAB(wl)) {
+		int i = 0;
+		for (i = 0; (atomic_read(&wl->callbacks) > callbacks) && i < 10000; i++) {
+			schedule();
+			flush_scheduled_work();
+		}
+	}
+	else
+#endif 
+	{
+
+		SPINWAIT((atomic_read(&wl->callbacks) > callbacks), 100 * 1000);
+	}
 
 	WL_LOCK(wl);
 }
@@ -1219,7 +1467,7 @@ wl_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
 			wl->pub->unit, cmd, preempt_count()));
 #endif
 
-#ifdef CONFIG_WIRELESS_EXT
+#ifdef USE_IW
 
 	if ((cmd >= SIOCIWFIRST) && (cmd <= SIOCIWLAST)) {
 
@@ -1308,7 +1556,7 @@ wl_get_stats(struct net_device *dev)
 	return (stats);
 }
 
-#ifdef CONFIG_WIRELESS_EXT
+#ifdef USE_IW
 struct iw_statistics *
 wl_get_wireless_stats(struct net_device *dev)
 {
@@ -1409,20 +1657,34 @@ wl_set_mac_address(struct net_device *dev, void *addr)
 static void
 wl_set_multicast_list(struct net_device *dev)
 {
-	_wl_set_multicast_list(dev);
+	if (!WL_ALL_PASSIVE_ENAB((wl_info_t *)WL_INFO(dev)))
+		_wl_set_multicast_list(dev);
+#ifdef WL_ALL_PASSIVE
+	else {
+		wl_info_t *wl = WL_INFO(dev);
+		wl->multicast_task.context = dev;
+
+		if (schedule_work(&wl->multicast_task.work)) {
+
+			atomic_inc(&wl->callbacks);
+		}
+	}
+#endif 
 }
 
 static void
 _wl_set_multicast_list(struct net_device *dev)
 {
-#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 33)
+#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 34)
 	struct dev_mc_list *mclist;
 #else
 	struct netdev_hw_addr   *ha;
 	int num;
 #endif
 	wl_info_t *wl;
-	int i;
+	int i, buflen;
+	struct maclist *maclist;
+	bool allmulti;
 
 	if (!dev)
 		return;
@@ -1430,38 +1692,49 @@ _wl_set_multicast_list(struct net_device *dev)
 
 	WL_TRACE(("wl%d: wl_set_multicast_list\n", wl->pub->unit));
 
-	WL_LOCK(wl);
-
 	if (wl->pub->up) {
-		wl->pub->allmulti = (dev->flags & IFF_ALLMULTI)? TRUE: FALSE;
+		allmulti = (dev->flags & IFF_ALLMULTI)? TRUE: FALSE;
 
-#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 33)
-		for (i = 0, mclist = dev->mc_list; mclist && (i < dev->mc_count);
-			i++, mclist = mclist->next) {
+		buflen = sizeof(struct maclist) + (MAXMULTILIST * ETHER_ADDR_LEN);
+
+		if ((maclist = MALLOC(wl->pub->osh, buflen)) == NULL) {
+			return;
+		}
+
+		i = 0;
+#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 34)
+		for (mclist = dev->mc_list; mclist && (i < dev->mc_count); mclist = mclist->next) {
 			if (i >= MAXMULTILIST) {
-				wl->pub->allmulti = TRUE;
+				allmulti = TRUE;
 				i = 0;
 				break;
 			}
-			wl->pub->multicast[i] = *((struct ether_addr*) mclist->dmi_addr);
+			bcopy(mclist->dmi_addr, &maclist->ea[i++], ETHER_ADDR_LEN);
 		}
 #else
 		num = min_t(int, netdev_mc_count(dev), MAXMULTILIST);
-		i = 0;
 		netdev_for_each_mc_addr(ha, dev) {
 			if (i >= num) {
-				wl->pub->allmulti = TRUE;
+				allmulti = TRUE;
 				i = 0;
 				break;
 			}
-			wl->pub->multicast[i] = *((struct ether_addr*) ha->addr);
+			bcopy(ha->addr,  &maclist->ea[i++], ETHER_ADDR_LEN);
 		}
 #endif 
-		wl->pub->nmulticast = i;
+		maclist->count = i;
+
+		WL_LOCK(wl);
+
+		wlc_iovar_setint(wl->wlc, "allmulti", allmulti);
 		wlc_set(wl->wlc, WLC_SET_PROMISC, (dev->flags & IFF_PROMISC));
+
+		wlc_iovar_op(wl->wlc, "mcast_list", NULL, 0, maclist, buflen, IOV_SET, NULL);
+
+		WL_UNLOCK(wl);
+		MFREE(wl->pub->osh, maclist, buflen);
 	}
 
-	WL_UNLOCK(wl);
 }
 
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 20)
@@ -1485,7 +1758,24 @@ wl_isr(int irq, void *dev_id, struct pt_regs *ptregs)
 		if (wantdpc) {
 
 			ASSERT(wl->resched == FALSE);
+#ifdef NAPI_POLL
+
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
+			napi_schedule(&wl->napi);
+#else
+			netif_rx_schedule(wl->dev);
+#endif 
+#else 
+#ifdef WL_ALL_PASSIVE
+			if (WL_ALL_PASSIVE_ENAB(wl)) {
+				if (schedule_work(&wl->wl_dpc_task.work))
+					atomic_inc(&wl->callbacks);
+				else
+					ASSERT(0);
+			} else
+#endif 
 			tasklet_schedule(&wl->tasklet);
+#endif 
 		}
 	}
 
@@ -1494,19 +1784,39 @@ wl_isr(int irq, void *dev_id, struct pt_regs *ptregs)
 	return IRQ_RETVAL(ours);
 }
 
+#ifdef NAPI_POLL
+static int BCMFASTPATH
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
+wl_poll(struct napi_struct *napi, int budget)
+#else
+wl_poll(struct net_device *dev, int *budget)
+#endif 
+#else 
 static void BCMFASTPATH
 wl_dpc(ulong data)
+#endif 
 {
 	wl_info_t *wl;
 
-	wl = (wl_info_t*) data;
+#ifdef NAPI_POLL
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
+	wl = (wl_info_t *)container_of(napi, wl_info_t, napi);
+	wl->pub->tunables->rxbnd = min(RXBND, budget);
+#else
+	wl = WL_INFO(dev);
+	wl->pub->tunables->rxbnd = min(RXBND, *budget);
+	ASSERT(wl->pub->tunables->rxbnd <= dev->quota);
+#endif 	
+#else 
+
+	wl = (wl_info_t *)data;
 
 	WL_LOCK(wl);
+#endif 
 
 	if (wl->pub->up) {
 		if (wl->resched) {
-			unsigned long flags;
-
+			unsigned long flags = 0;
 			INT_LOCK(wl, flags);
 			wlc_intrsupd(wl->wlc);
 			INT_UNLOCK(wl, flags);
@@ -1515,18 +1825,101 @@ wl_dpc(ulong data)
 		wl->resched = wlc_dpc(wl->wlc, TRUE);
 	}
 
-	if (!wl->pub->up)
-		goto done;
+	if (!wl->pub->up) {
+#ifdef WL_ALL_PASSIVE
 
+		if ((WL_ALL_PASSIVE_ENAB(wl))) {
+			atomic_dec(&wl->callbacks);
+		}
+#endif 
+		goto done;
+	}
+
+#ifndef NAPI_POLL
+#ifdef WL_ALL_PASSIVE
+	if (wl->resched) {
+		if (!(WL_ALL_PASSIVE_ENAB(wl)))
+			tasklet_schedule(&wl->tasklet);
+		else
+			if (!schedule_work(&wl->wl_dpc_task.work))
+
+				ASSERT(0);
+	}
+	else {
+
+		if (WL_ALL_PASSIVE_ENAB(wl))
+			atomic_dec(&wl->callbacks);
+		wl_intrson(wl);
+	}
+#else 
 	if (wl->resched)
 		tasklet_schedule(&wl->tasklet);
 	else {
 
 		wl_intrson(wl);
 	}
+#endif 
 
 done:
 	WL_UNLOCK(wl);
+	return;
+#else 
+
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
+	WL_TRACE(("wl%d: wl_poll: rxbnd %d budget %d processed %d\n",
+		wl->pub->unit, wl->pub->rxbnd, budget, wl->pub->processed));
+
+	ASSERT(wl->pub->processed <= wl->pub->tunables->rxbnd);
+
+	if (!wl->resched) {
+		napi_complete(&wl->napi);
+
+		wl_intrson(wl);
+	}
+	return wl->pub->processed;
+done:
+	return 0;
+
+#else 
+	WL_TRACE(("wl%d: wl_poll: rxbnd %d quota %d budget %d processed %d\n",
+	          wl->pub->unit, wl->pub->rxbnd, dev->quota,
+	          *budget, wl->pub->processed));
+
+	ASSERT(wl->pub->processed <= wl->pub->tunables->rxbnd);
+
+	*budget -= wl->pub->processed;
+	dev->quota -= wl->pub->processed;
+
+	if (wl->resched)
+
+		return 1;
+
+	netif_rx_complete(dev);
+
+	wl_intrson(wl);
+done:
+	return 0;
+#endif 
+#endif 
+}
+
+#if defined(WL_ALL_PASSIVE)
+static void
+wl_dpc_rxwork(struct wl_task *task)
+{
+	wl_info_t *wl = (wl_info_t *)task->context;
+	WL_TRACE(("wl%d: %s\n", wl->pub->unit, __FUNCTION__));
+
+	wl_dpc((unsigned long)wl);
+	return;
+}
+#endif 
+
+static inline int32
+wl_ctf_forward(wl_info_t *wl, struct sk_buff *skb)
+{
+
+	return (BCME_ERROR);
 }
 
 void BCMFASTPATH
@@ -1552,7 +1945,7 @@ wl_sendup(wl_info_t *wl, wl_if_t *wlif, void *p, int numpkt)
 		skb->dev = wl->dev;
 	}
 
-	skb->protocol = eth_type_trans(skb, wl->dev);
+	skb->protocol = eth_type_trans(skb, skb->dev);
 	if (!ISALIGNED((uintptr)skb->data, 4)) {
 		WL_ERROR(("Unaligned assert. skb %p. skb->data %p.\n", skb, skb->data));
 		if (wlif) {
@@ -1568,7 +1961,11 @@ wl_sendup(wl_info_t *wl, wl_if_t *wlif, void *p, int numpkt)
 	WL_APSTA_RX(("wl%d: wl_sendup(): pkt %p summed %d on interface %p (%s)\n",
 		wl->pub->unit, p, skb->ip_summed, wlif, skb->dev->name));
 
+#ifdef NAPI_POLL
+	netif_receive_skb(skb);
+#else 
 	netif_rx(skb);
+#endif 
 }
 
 void
@@ -1624,7 +2021,7 @@ wl_link_down(wl_info_t *wl, char *ifname)
 void
 wl_event(wl_info_t *wl, char *ifname, wlc_event_t *e)
 {
-#ifdef CONFIG_WIRELESS_EXT
+#ifdef USE_IW
 	wl_iw_event(wl->dev, &(e->event), e->data);
 #endif 
 
@@ -1636,7 +2033,7 @@ wl_event(wl_info_t *wl, char *ifname, wlc_event_t *e)
 		else
 			wl_link_down(wl, ifname);
 		break;
-#if defined(WL_CONFIG_RFKILL_INPUT)
+#if defined(WL_CONFIG_RFKILL)
 	case WLC_E_RADIO: {
 		mbool i;
 		if (wlc_get(wl->wlc, WLC_GET_RADIO, &i) < 0)
@@ -1647,7 +2044,7 @@ wl_event(wl_info_t *wl, char *ifname, wlc_event_t *e)
 		wl->last_phyind = (mbool)(i & WL_RADIO_HW_DISABLE);
 
 		WL_ERROR(("wl%d: Radio hardware state changed to %d\n", wl->pub->unit, i));
-		(void) wl_schedule_task(wl, wl_force_kill, wl);
+		wl_report_radio_state(wl);
 		break;
 	}
 #else
@@ -1662,32 +2059,180 @@ wl_event_sync(wl_info_t *wl, char *ifname, wlc_event_t *e)
 {
 }
 
+#ifndef WL_THREAD
+static int BCMFASTPATH
+wl_start(struct sk_buff *skb, struct net_device *dev)
+{
+	wl_if_t *wlif;
+	wl_info_t *wl;
+
+	if (!dev)
+		return -ENETDOWN;
+
+	wlif = WL_DEV_IF(dev);
+	wl = WL_INFO(dev);
+
+	if (!WL_ALL_PASSIVE_ENAB(wl))
+		return wl_start_int(wl, wlif, skb);
+#ifdef WL_ALL_PASSIVE
+	else {
+		skb->prev = NULL;
+
+		TXQ_LOCK(wl);
+		if (wl->txq_head == NULL)
+			wl->txq_head = skb;
+		else {
+			wl->txq_tail->prev = skb;
+		}
+		wl->txq_tail = skb;
+
+		if (wl->txq_dispatched == FALSE) {
+			wl->txq_dispatched = TRUE;
+
+			if (schedule_work(&wl->txq_task.work)) {
+				atomic_inc(&wl->callbacks);
+			} else {
+				WL_ERROR(("wl%d: wl_start/schedule_work failed\n", wl->pub->unit));
+				wl->txq_dispatched = FALSE;
+			}
+		}
+
+		TXQ_UNLOCK(wl);
+	}
+#endif 
+
+	return (0);
+}
+#endif 
+
+#ifdef WL_ALL_PASSIVE
+static void
+wl_start_txqwork(struct wl_task *task)
+{
+	wl_info_t *wl = (wl_info_t *)task->context;
+	struct sk_buff *skb;
+	uint count = 0;
+
+	WL_TRACE(("wl%d: wl_start_txqwork\n", wl->pub->unit));
+
+	TXQ_LOCK(wl);
+	while (wl->txq_head) {
+		skb = wl->txq_head;
+		wl->txq_head = skb->prev;
+		skb->prev = NULL;
+		if (wl->txq_head == NULL)
+			wl->txq_tail = NULL;
+		TXQ_UNLOCK(wl);
+
+		wl_start_int(wl, WL_DEV_IF(skb->dev), skb);
+
+		if (++count >= 10)
+			break;
+
+		TXQ_LOCK(wl);
+	}
+
+	if (count >= 10) {
+		if (!schedule_work(&wl->txq_task.work)) {
+			WL_ERROR(("wl%d: wl_start/schedule_work failed\n", wl->pub->unit));
+			atomic_dec(&wl->callbacks);
+		}
+	} else {
+		wl->txq_dispatched = FALSE;
+		TXQ_UNLOCK(wl);
+		atomic_dec(&wl->callbacks);
+	}
+
+	return;
+}
+
+static void
+wl_txq_free(wl_info_t *wl)
+{
+	struct sk_buff *skb;
+
+	if (wl->txq_head == NULL) {
+		ASSERT(wl->txq_tail == NULL);
+		return;
+	}
+
+	while (wl->txq_head) {
+		skb = wl->txq_head;
+		wl->txq_head = skb->prev;
+		PKTFREE(wl->osh, skb, TRUE);
+	}
+
+	wl->txq_tail = NULL;
+}
+#endif 
+
+#ifdef WL_ALL_PASSIVE
+static void
+wl_set_multicast_list_workitem(struct work_struct *work)
+{
+	wl_task_t *task = (wl_task_t *)work;
+	struct net_device *dev = (struct net_device*)task->context;
+	wl_info_t *wl;
+
+	wl = WL_INFO(dev);
+
+	atomic_dec(&wl->callbacks);
+
+	_wl_set_multicast_list(dev);
+}
+
+static void
+wl_timer_task(wl_task_t *task)
+{
+	wl_timer_t *t = (wl_timer_t *)task->context;
+
+	_wl_timer(t);
+	MFREE(t->wl->osh, task, sizeof(wl_task_t));
+
+	atomic_dec(&t->wl->callbacks);
+}
+#endif 
+
 static void
 wl_timer(ulong data)
 {
-	_wl_timer((wl_timer_t*)data);
+	wl_timer_t *t = (wl_timer_t *)data;
+
+	if (!WL_ALL_PASSIVE_ENAB(t->wl))
+		_wl_timer(t);
+#ifdef WL_ALL_PASSIVE
+	else
+		wl_schedule_task(t->wl, wl_timer_task, t);
+#endif 	
 }
 
 static void
 _wl_timer(wl_timer_t *t)
 {
-	WL_LOCK(t->wl);
+	wl_info_t *wl = t->wl;
 
-	if (t->set) {
+	WL_LOCK(wl);
+
+	if (t->set && (!timer_pending(&t->timer))) {
 		if (t->periodic) {
 			t->timer.expires = jiffies + t->ms*HZ/1000;
-			atomic_inc(&t->wl->callbacks);
+			atomic_inc(&wl->callbacks);
 			add_timer(&t->timer);
 			t->set = TRUE;
 		} else
 			t->set = FALSE;
 
 		t->fn(t->arg);
+#ifdef BCMDBG
+		wlc_update_perf_stats(wl->wlc, WLC_PERF_STATS_TMR_DPC);
+		t->ticks++;
+#endif
+
 	}
 
-	atomic_dec(&t->wl->callbacks);
+	atomic_dec(&wl->callbacks);
 
-	WL_UNLOCK(t->wl);
+	WL_UNLOCK(wl);
 }
 
 wl_timer_t *
@@ -1695,9 +2240,11 @@ wl_init_timer(wl_info_t *wl, void (*fn)(void *arg), void *arg, const char *name)
 {
 	wl_timer_t *t;
 
-	if (!(t = MALLOC(wl->osh, sizeof(wl_timer_t)))) {
-		WL_ERROR(("wl%d: wl_init_timer: out of memory, malloced %d bytes\n", wl->pub->unit,
-			MALLOCED(wl->osh)));
+	t = (wl_timer_t*)MALLOC(wl->osh, sizeof(wl_timer_t));
+
+	if (t == NULL) {
+		WL_ERROR(("wl%d: wl_init_timer: out of memory, malloced %d bytes\n",
+		          wl->unit, MALLOCED(wl->osh)));
 		return 0;
 	}
 
@@ -1729,10 +2276,13 @@ wl_add_timer(wl_info_t *wl, wl_timer_t *t, uint ms, int periodic)
 			__FUNCTION__, t->name, periodic));
 	}
 #endif
-	ASSERT(!t->set);
 
 	t->ms = ms;
 	t->periodic = (bool) periodic;
+
+	if (t->set)
+		return;
+
 	t->set = TRUE;
 	t->timer.expires = jiffies + ms*HZ/1000;
 
@@ -1743,11 +2293,12 @@ wl_add_timer(wl_info_t *wl, wl_timer_t *t, uint ms, int periodic)
 bool
 wl_del_timer(wl_info_t *wl, wl_timer_t *t)
 {
+	ASSERT(t);
 	if (t->set) {
 		t->set = FALSE;
 		if (!del_timer(&t->timer)) {
 #ifdef BCMDBG
-			WL_INFORM(("wl%d: Failed to delete timer %s\n", wl->pub->unit, t->name));
+			WL_INFORM(("wl%d: Failed to delete timer %s\n", wl->unit, t->name));
 #endif
 			return FALSE;
 		}
@@ -2044,71 +2595,62 @@ wl_tkip_printstats(wl_info_t *wl, bool group_key)
 #endif 
 }
 
-#if defined(WL_CONFIG_RFKILL_INPUT)
+#if defined(WL_CONFIG_RFKILL)   
 
 static int
-wl_sw_toggle_radio(void *data, enum rfkill_state state)
+wl_set_radio_block(void *data, bool blocked)
 {
 	wl_info_t *wl = data;
-	mbool cur_state;
-	uint32 radioval = WL_RADIO_SW_DISABLE << 16;
+	uint32 radioval;
 
-	if (wlc_get(wl->wlc, WLC_GET_RADIO, &cur_state) < 0) {
-		WL_ERROR(("%s: WLC_GET_RADIO failed\n", __FUNCTION__));
+	radioval = WL_RADIO_SW_DISABLE << 16 | blocked;
+
+	WL_LOCK(wl);
+
+	if (wlc_set(wl->wlc, WLC_SET_RADIO, radioval) < 0) {
+		WL_ERROR(("%s: SET_RADIO failed\n", __FUNCTION__));
+		return 1;
 	}
 
-	switch (state) {
-	case RFKILL_STATE_SOFT_BLOCKED:
+	WL_UNLOCK(wl);
 
-		if (!mboolisset(cur_state, WL_RADIO_SW_DISABLE)) {
-			WL_ERROR(("%s: => Disable radio\n", __FUNCTION__));
-		}
-		return 0;
-	case RFKILL_STATE_UNBLOCKED:
-
-		if (mboolisset(cur_state, WL_RADIO_SW_DISABLE)) {
-			radioval |= WL_RADIO_SW_DISABLE;
-			WL_ERROR(("%s: => Enable radio\n", __FUNCTION__));
-		}
-		return 0;
-	default:
-		return -EINVAL;
-	}
-	if (wlc_set(wl->wlc, WLC_SET_RADIO, radioval) < 0)
-			WL_ERROR(("%s: SET_RADIO failed\n", __FUNCTION__));
+	return 0;
 }
 
+static const struct rfkill_ops bcmwl_rfkill_ops = {
+	.set_block = wl_set_radio_block,
+};
+
 static int
 wl_init_rfkill(wl_info_t *wl)
 {
 	int status;
-	wl->wl_rfkill.rfkill = rfkill_allocate(&wl->dev->dev, RFKILL_TYPE_WLAN);
+
+	snprintf(wl->wl_rfkill.rfkill_name, sizeof(wl->wl_rfkill.rfkill_name),
+	"brcmwl-%d", wl->pub->unit);
+
+	wl->wl_rfkill.rfkill = rfkill_alloc(wl->wl_rfkill.rfkill_name, &wl->dev->dev,
+	RFKILL_TYPE_WLAN, &bcmwl_rfkill_ops, wl);
+
 	if (!wl->wl_rfkill.rfkill) {
 		WL_ERROR(("%s: RFKILL: Failed to allocate rfkill\n", __FUNCTION__));
 		return -ENOMEM;
 	}
 
-	snprintf(wl->wl_rfkill.rfkill_name, sizeof(wl->wl_rfkill.rfkill_name),
-		"brcmwl-%d", wl->pub->unit);
-
-	if (wlc_get(wl->wlc, WLC_GET_RADIO, &status) < 0)
+	if (wlc_get(wl->wlc, WLC_GET_RADIO, &status) < 0) {
 		WL_ERROR(("%s: WLC_GET_RADIO failed\n", __FUNCTION__));
+		return 1;
+	}
 
-	wl->wl_rfkill.rfkill->name = wl->wl_rfkill.rfkill_name;
-	wl->wl_rfkill.rfkill->data = wl;
-	wl->wl_rfkill.rfkill->toggle_radio = wl_sw_toggle_radio;
-
-	if (status & WL_RADIO_HW_DISABLE)
-		wl->wl_rfkill.rfkill->state = RFKILL_STATE_HARD_BLOCKED;
-	else
-		wl->wl_rfkill.rfkill->state = RFKILL_STATE_UNBLOCKED;
-
-	wl->wl_rfkill.rfkill->user_claim_unsupported = 1;
+	rfkill_init_sw_state(wl->wl_rfkill.rfkill, status);
 
 	if (rfkill_register(wl->wl_rfkill.rfkill)) {
 		WL_ERROR(("%s: rfkill_register failed! \n", __FUNCTION__));
-		rfkill_free(wl->wl_rfkill.rfkill);
+		rfkill_destroy(wl->wl_rfkill.rfkill);
+		return 2;
 	}
+
+	WL_ERROR(("%s: rfkill registered\n", __FUNCTION__));
 	wl->wl_rfkill.registered = TRUE;
 	return 0;
 }
@@ -2118,25 +2660,19 @@ wl_uninit_rfkill(wl_info_t *wl)
 {
 	if (wl->wl_rfkill.registered) {
 		rfkill_unregister(wl->wl_rfkill.rfkill);
+		rfkill_destroy(wl->wl_rfkill.rfkill);
 		wl->wl_rfkill.registered = FALSE;
 		wl->wl_rfkill.rfkill = NULL;
 	}
 }
 
 static void
-wl_force_kill(wl_task_t *task)
+wl_report_radio_state(wl_info_t *wl)
 {
-	wl_info_t *wl = (wl_info_t *) task->context;
 
-	ASSERT(wl);
-	if (wl->last_phyind)
-		rfkill_force_state(wl->wl_rfkill.rfkill, RFKILL_STATE_HARD_BLOCKED);
-	else
-		rfkill_force_state(wl->wl_rfkill.rfkill, RFKILL_STATE_UNBLOCKED);
-
-	MFREE(wl->osh, task, sizeof(wl_task_t));
-	atomic_dec(&wl->callbacks);
+	rfkill_set_hw_state(wl->wl_rfkill.rfkill, wl->last_phyind != 0);
 }
+
 #endif 
 
 static int
@@ -2145,7 +2681,7 @@ wl_linux_watchdog(void *ctx)
 	wl_info_t *wl = (wl_info_t *) ctx;
 	struct net_device_stats *stats = NULL;
 	uint id;
-#ifdef CONFIG_WIRELESS_EXT
+#ifdef USE_IW
 	struct iw_statistics *wstats = NULL;
 	int phy_noise;
 #endif
@@ -2156,7 +2692,7 @@ wl_linux_watchdog(void *ctx)
 		id = 1 - wl->stats_id;
 
 		stats = &wl->stats_watchdog[id];
-#ifdef CONFIG_WIRELESS_EXT
+#ifdef USE_IW
 		wstats = &wl->wstats_watchdog[id];
 #endif
 		stats->rx_packets = WLCNTVAL(wl->pub->_cnt->rxframe);
@@ -2175,7 +2711,7 @@ wl_linux_watchdog(void *ctx)
 		stats->rx_missed_errors = 0;
 
 		stats->tx_fifo_errors = WLCNTVAL(wl->pub->_cnt->txuflo);
-#ifdef CONFIG_WIRELESS_EXT
+#ifdef USE_IW
 #if WIRELESS_EXT > 11
 		wstats->discard.nwid = 0;
 		wstats->discard.code = WLCNTVAL(wl->pub->_cnt->rxundec);
@@ -2190,7 +2726,7 @@ wl_linux_watchdog(void *ctx)
 
 		wl->stats_id = id;
 
-#ifdef CONFIG_WIRELESS_EXT
+#ifdef USE_IW
 		if (!wlc_get(wl->wlc, WLC_GET_PHY_NOISE, &phy_noise))
 			wl->phy_noise = phy_noise;
 #endif 
